/*
 *  child.c
 *
 *  Timo.Felbinger@quantum.physik.uni-potsdam.de, 1999
 *
 *  functions to work with child processes
 *
 *  Note that some related low-level functions are in os_unix.c
 *  Some user call-able functions are in eval.c
 *
 */


#include "vim.h"

#ifdef USE_CHILD


  CHILD *
find_child_by_name( buf, name )
  BUF * buf;
  char_u * name;
{
  CHILD *p;
  for( p = buf->b_child; p ; p = p->c_next ) 
    if( !strcmp( name, p->c_name ) )
      return p;
  return NULL;
}


  void
kill_all_childs( buf )
  BUF *buf;
{
  while( buf->b_child )
    kill_child( buf->b_child );
}

  int
list_childs( buf )
  BUF *buf;
{
  CHILD *pc;
  char_u mbuf[ 64 ];

  if( ! buf->b_child ) {
    MSG( "buffer has no childs" );
    return FALSE;
  }
  for( pc = buf->b_child; pc; pc = pc->c_next ) {
#ifdef HAVE_SNPRINTF
    snprintf( 
      mbuf, 64
#else
  /*
   *  Those with ancient systems get what they deserve:
   */
    sprintf( 
      mbuf
#endif
    , "  %6ld:  %4di %4do %4de %4da%c  %s"
    , (long)(pc->c_pid), pc->c_in, pc->c_out, pc->c_err, pc->c_async 
    , (pc->c_handler ? '*' : '-' ), pc->c_name
    );
    mbuf[63] = NUL;
    MSG( mbuf );
  }
  return TRUE;
}

/*
 *  child_read_async(): read input from child's async channel.
 *  this will block if no data is available, so a select() should be
 *  used before calling this function.
 *  return TRUE, iff any data was passed to the handler (this happens
 *  only after a newline is read, so we may actually read some data and
 *  nevertheless return FALSE).
 *
 *  this function is called from within the low-level input function
 *  RealWaitForChar (and maybe others?).
 *
 */
  int
child_read_async( pc )
  CHILD *pc;
{
  ssize_t rsize;
  char_u *newend;
  char_u *newstart;
  int did_handle = FALSE;

  if( pc->c_be - pc->c_bp < CHILD_ASYNC_BSIZE ) {
    char_u *newp = (char_u *)realloc( pc->c_inbuf, 2 * (pc->c_be - pc->c_inbuf) );
    if( !newp ) {
      EMSG( "child_read_async(): realloc() failed" );
      return FALSE;
    }
    pc->c_bp = (pc->c_bp - pc->c_inbuf) + newp;
    pc->c_be = newp + 2 * (pc->c_be - pc->c_inbuf);
    pc->c_inbuf = newp;
  }
  errno = 0;
  rsize = read( pc->c_async, pc->c_bp, pc->c_be - pc->c_bp );
  if( errno || (rsize <= 0) ) {
    EMSG( "child_read_async(): read() failed, disabling async input" );
    pc->c_handler = NULL;
    close( pc->c_async );
    pc->c_async = -1;
    return FALSE;
  }
  if( rsize == 0 ) {
    EMSG( "child_read_async(): EOF reached, disabling async input" );
    pc->c_handler = NULL;
    close( pc->c_async );
    pc->c_async = -1;
    return FALSE;
  }
  newstart = pc->c_inbuf;  /* pointer to not-handled input */
  newend = pc->c_bp + rsize;  /* end of input data */
  while( pc->c_bp < newend ) {
    if( *(pc->c_bp) == '\n' ) {   /* we have a complete line and call the handler */
      *(pc->c_bp) = NUL;
      pc->c_handler( pc, newstart );
      did_handle = TRUE;
      newstart = pc->c_bp+1;
    }
    ++(pc->c_bp);
  }
  if( newstart > pc->c_inbuf ) { /* shift remaining data to start of buffer */
    char_u *tmpp;
    pc->c_bp -= (newstart - pc->c_inbuf);
    for( tmpp = pc->c_inbuf; newstart < newend; )
      *( tmpp++ ) = *( newstart++ );
  }
  return did_handle;
}

/*
 *  child_handle_emsg:
 *  this function handles error messages which have been read
 *  from a child's stderr
 */
  static void
child_handle_emsg( pc, msg )
  CHILD *pc;
  char_u *msg;
{
  char_u *fmsg;
  if( ! ( fmsg = (char_u *)malloc( strlen(msg) + 32 + strlen(pc->c_name) ) ) ) {
    EMSG( "child_handle_emsg: malloc() failed" );
    return;
  }
  sprintf( fmsg, "error in child %s: %s", pc->c_name, msg );
  EMSG( fmsg );
  free( fmsg );
}

/*
 *  child_read_stderr:
 *  similar to child_read_async, this function reads any incoming data 
 *  from a child's stderr. complete lines are assumed to be human-readable
 *  error messages and are passed to child_handle_emsg().
 */
  void
child_read_stderr( pc )
  CHILD *pc;
{
  ssize_t rsize;
  char_u *newend;
  char_u *newstart;

  if( pc->c_errbe - pc->c_errbp < CHILD_ASYNC_BSIZE ) {
    char_u *newp = (char_u *)realloc( pc->c_errbuf, 2 * (pc->c_errbe - pc->c_errbuf) );
    if( !newp ) {
      EMSG( "child_read_stderr(): realloc() failed" );
      return;
    }
    pc->c_errbp = (pc->c_errbp - pc->c_errbuf) + newp;
    pc->c_errbe = newp + 2 * (pc->c_errbe - pc->c_errbuf);
    pc->c_errbuf = newp;
  }
  errno = 0;
  rsize = read( pc->c_err, pc->c_errbp, pc->c_errbe - pc->c_errbp );
  if( errno || rsize < 0 ) {
    /*  in case of real failure, we close the stderr so we are not
     *  flooded with error messages. the child is kept alive, however,
     *  so that the user might still have a chance to save data.
     */
    EMSG( "child_read_stderr(): read() failed on child's stderr, disconnecting..." );
    close( pc->c_err );
    pc->c_err = -1;
    errno = 0;
    return;
  }
  if( rsize == 0 ) {
    EMSG( "EOF reached on child's stderr, disconnecting..." );
    close( pc->c_err );
    pc->c_err = -1;
    return;
  }
  newstart = pc->c_errbuf;  /* pointer to not-handled input */
  newend = pc->c_errbp + rsize;  /* end of input data */
  while( pc->c_errbp < newend ) {
    if( *(pc->c_errbp) == '\n' ) {   /* we have a complete line and call the handler */
      *(pc->c_errbp) = NUL;
      child_handle_emsg( pc, newstart );
      newstart = pc->c_errbp + 1;
    }
    ++(pc->c_errbp);
  }
  if( newstart > pc->c_errbuf ) { /* shift remaining data to start of buffer */
    char_u *tmpp;
    pc->c_errbp -= (newstart - pc->c_errbuf);
    for( tmpp = pc->c_errbuf; newstart < newend; )
      *( tmpp++ ) = *( newstart++ );
  }
}


/*
 *  some functions which should actually be somewhere else, as they are
 *  not really related to childs. However, I didn't find them anywhere,
 *  so I had to write them.
 *  they are pretty much copied from the perl interface in if_perl.xs
 */

 int
child_set_line( buf, lnum, line )
  BUF *buf;
  linenr_t  lnum;
  char_u   *line;
{
  int rv = FAIL;
  BUF *savebuf = curbuf;

  if( lnum < 1 || lnum > buf->b_ml.ml_line_count )
    return rv;

  curbuf = buf;
  if( u_savesub(lnum) ) {
    rv = ml_replace( lnum, line, TRUE );
    changed();
#ifdef SYNTAX_HL
    /* recompute syntax hl. for this line */
    syn_changed(lnum);
#endif
  }
  curbuf = savebuf;
  update_curbuf(NOT_VALID);

  c_need_refresh = TRUE;

  return rv;
}

  int
child_append_line( buf, lnum, line, llen )
  BUF *buf;
  linenr_t  lnum;
  char_u   *line;
  colnr_t   llen;   /* length _including_ the NUL */
{
  int rv = FAIL;
  BUF *savebuf = curbuf;
  WIN *win;

  if( lnum < 0 || lnum > buf->b_ml.ml_line_count )
    return rv;

  curbuf = buf;
  if( u_inssub( lnum + 1 ) ) {
    mark_adjust( lnum+1, MAXLNUM, 1L, 0L );
    ml_append( lnum, line, llen, FALSE );
    changed();
    rv = TRUE;
  }
  curbuf = savebuf;
  update_curbuf(NOT_VALID);

  /* fix cursor position: */
  for( win = firstwin; win; win = win->w_next )
    if( win->w_buffer == buf )
      if( win->w_cursor.lnum > lnum )
        win->w_cursor.lnum++;

  c_need_refresh = TRUE;

  return rv;
}

  int
child_delete_line( buf, lnum )
  BUF *buf;
  linenr_t lnum;
{
  int rv = FAIL;
  BUF *savebuf = curbuf;
  WIN *win;

  if( lnum < 1 || lnum > buf->b_ml.ml_line_count )
    return rv;

  curbuf = buf;
  if( u_savedel( lnum , 1 ) ) {
    mark_adjust( lnum, lnum, MAXLNUM, -1 );
    rv = ml_delete( lnum, 0 );
    changed();
  }
  curbuf = savebuf;
  update_curbuf(NOT_VALID);

  /* fix cursor position: */
  for( win = firstwin; win; win = win->w_next )
    if( win->w_buffer == buf )
      if( win->w_cursor.lnum >= lnum )
        win->w_cursor.lnum--;

  c_need_refresh = TRUE;

  return rv;
}

  char_u *
child_get_line( buf, lnum )
  BUF *buf;
  linenr_t lnum;
{
  if( lnum < 1 || lnum > buf->b_ml.ml_line_count )
    return NULL;
  return ml_get_buf( buf, lnum, FALSE );
}


#endif  /* defined( USE_CHILD ) */

