/*
 *  if_mma.c
 *
 *  Copyright (C) 1999 Timo Felbinger
 *  email:  Timo.Felbinger@quantum.physik.uni-potsdam.de
 *
 *  Interface  Vim  <---->   Mathematica
 *
 *  This requires a small auxiliary program, ml2vim, which basically
 *  translates mathlink to ASCII, which is available as a separate
 *  package. ml2vim must be fork()ed as a child process, so you must
 *  compile vim with --enable-child to use this interface.
 *
 *  This version speaks the same protocol as ml2vim-0.99.8
 *
 *  The first part of this file contains functions which are used to
 *  send input to mma for evaluation; the second part deals with the
 *  reply received from mma.
 *
 */

#include "vim.h"

#ifdef USE_MMA

#define MAX_TAG_LEN 64


#ifdef HAVE_SNPRINTF
#  define SNPRINTF_EXTRA_ARG(X) X,
#else
  /*
   *  those with very ancient systems get what they deserve:
   */
#  define snprintf sprintf
#  define SNPRINTF_EXTRA_ARG(X)
#endif

#undef DEBUG

#ifdef DEBUG
# include <stdarg.h>
  FILE *stdlog = NULL;

  void utc_date_time( char *s, time_t ctm ) {
    struct tm *ptm;

    ptm = gmtime( &ctm );
    snprintf( 
      s, SNPRINTF_EXTRA_ARG( 16 ) "%04u%02u%02u.%02u%02u%02u"
      , 1900 + ptm->tm_year, (ptm->tm_mon)+1, ptm->tm_mday
      , ptm->tm_hour, ptm->tm_min, ptm->tm_sec
    );
  }
  char *utc_now( char *s ) {
    utc_date_time( s, time(NULL) );
    return s;
  }

  void trace ( char *fmt, ... ) {
    char now[16];
    va_list vap;
    va_start( vap, fmt );
    if( ! stdlog )
      stdlog = fopen( "if_mma.log", "a" );
    if( stdlog ) {
      utc_now( now );
      fprintf( stdlog, "\n%s: ", now );
      vfprintf( stdlog, fmt, vap );
      if( errno )
        fprintf( stdlog, "  errno: %d", errno );
      fflush( stdlog );
    }
    va_end( vap );
  }
#else
  void trace( char *fmt, ... ) { }
#endif


/*
 *
 *   functions to parse input lines:
 *
 */


/*
 *  is_separator_line: note that the result of this function is
 *  only correct for lines which don't belong to a block!
 */
  static int
mma_is_separator_line( lp )
  char_u *lp;
{
  if( ! lp )
    return FALSE;
  while( vim_iswhite( *lp ) )
    ++lp;
  if( *lp == '-' || *lp == '=' ) {
    ++lp;
    if( *lp == '-' || *lp == '=' ) {
      ++lp;
      return *lp == '-' || *lp == '=';
    }
  }
  return FALSE;
}

/*
 *   We allow comments inside blocks, and also between option lines.
 *   They may start with either # or % and must be indented at most as
 *   far as the input tag. This indentation rule is required, as there
 *   seems to be no good choice for a simple comment syntax which could
 *   not legally occur in normal Mma input.
 */
  static int
mma_is_comment_line( lp )
  char_u *lp;
{
  if( ! lp )
    return FALSE;
  while( vim_iswhite( *lp ) )
    ++lp;
  return  *lp == '%' || *lp == '#';
}


  static int
mma_is_option_line( lp )
  char_u *lp;
{
  if( ! lp )
    return FALSE;
  while( vim_iswhite( *lp ) )
    ++lp;
  if( *lp != '-' )      /*  flag:  is option line  */
    return FALSE;
  if( * ++lp == '-' )   /*  flag:  long option  */
    ++lp;
  if( *lp == '!' )      /*  flag:  global option  */
    ++lp;
  if( *lp == '/' )      /*  flag:  invert option  */
    ++lp;
  return isalnum( *lp );  /*  now we need something alphanumeric  */
}


/*  possible states of a tagged line:  */

#define MMA_TAG_VIRGIN    1    /*  single tag number: 0  */
#define MMA_TAG_SUBMITTED 2    /*  submitted, but no results received  */
#define MMA_TAG_PARTIAL   3    /*  in work and some results been received  */
#define MMA_TAG_EVALUATED 4    /*  completely evaluated, all results received  */
#define MMA_TAG_MALFORMED 5    /*  not a well-formed tagged line  */

struct mma_tline_s {
  long int vim_first;   /*  first of a range of _negative_ numbers    */
  long int vim_last;
  long int mma_first;   /*  first of a range of _positive_ numbers    */
  long int mma_last;
  char_u   *options;     /*  pointer to options, or "" iff no options  */
  char_u   *payload;
};

/*
 *  analyze a tagged line:
 *  This will return the status of a tagged line (one of the above),
 *  and fill in all applicable entries in the structure passed to it
 *  by reference. All non-applicable entries are left undefined:
 *  you _must_ check the return value before using one!
 *  This is to make this function efficient: it will exit as soon as possible
 *  for a malformed line, so that it can be used to check large sets of lines
 *  inexpensively.
 *
 *  This function is the authoritative definition of a tagged line in a .vma file.
 */
  static int
mma_analyze_tagged_line( lp, presult )
  char_u *lp;
  struct mma_tline_s *presult;
{
  long int first_num, last_num;
  int neg;
  int type;

  trace( "mma_analyze_tagged_line: >>>%s<<<", lp );

  if( ! lp )
    return MMA_TAG_MALFORMED;
  while( vim_iswhite( *lp ) )
    ++lp;
  if( ! isalpha( *lp ) )
    return MMA_TAG_MALFORMED;
  while( isalnum( *lp ) || *lp == '_' )
    ++lp;
  if( *lp++ != '[' )
    return MMA_TAG_MALFORMED;

  trace( "  mma_analyze_tagged_line: found `['" );

  if( ( neg = ( *lp == '-' ) ) )
    ++lp;
  if( ! isdigit( *lp ) )
    return MMA_TAG_MALFORMED;
  for( first_num = 0; isdigit( *lp ); ++lp )
    first_num = 10 * first_num + ( *lp - (char_u)'0' );
  if( neg ) first_num = -first_num;

  if( *lp == '.' ) {
    if( lp[1] == '.' && lp[2] == '.' ) {
      trace( "  found 3 dots (first range)" );
      lp += 3;
      /*  if( ! first_num ) return MMA_TAG_MALFORMED;  */
      if( ( neg = ( *lp == '-' ) ) )
        ++lp;
      if( ! isdigit( *lp ) )
        return MMA_TAG_MALFORMED;
      for( last_num = 0; isdigit( *lp ); ++lp )
        last_num = 10 * last_num + ( *lp - (char_u)'0' );
      if( neg ) last_num = -last_num;
    } else {
      return MMA_TAG_MALFORMED;
    }
  } else {
    last_num = first_num;
  }

  if( *lp == ',' ) {
    ++lp;
    trace( "  found comma, parsing second range" );

    if( first_num >= 0 || last_num > first_num )
      return MMA_TAG_MALFORMED;

    presult->vim_first = first_num;
    presult->vim_last = last_num;
    type = MMA_TAG_PARTIAL;

    if( ! isdigit( *lp ) )
      return MMA_TAG_MALFORMED;
    for( first_num = 0; isdigit( *lp ); ++lp )
      first_num = 10 * first_num + ( *lp - (char_u)'0' );
    if( ! first_num )
      return MMA_TAG_MALFORMED;

    if( *lp == '.' ) {
      if( lp[1] == '.' && lp[2] == '.' ) {
        trace( "  found 3 dots (second range)" );
        lp += 3;
        if( ! isdigit( *lp ) )
          return MMA_TAG_MALFORMED;
        for( last_num = 0; isdigit( *lp ); ++lp )
          last_num = 10 * last_num + ( *lp - (char_u)'0' );
      } else {
        return MMA_TAG_MALFORMED;
      }
    } else {
      last_num = first_num;
    }

    presult->mma_first = first_num;
    presult->mma_last = last_num;

    trace( 
      "mma_analyze_tagged_line: PARTIAL: %ld...%ld,%ld...%ld"
    , presult->vim_first , presult->vim_last
    , presult->mma_first , presult->mma_last
    );

  } else {
    if( first_num < 0 && last_num <= first_num ) {
      presult->vim_first = first_num;
      presult->vim_last = last_num;
      type = MMA_TAG_SUBMITTED;
      trace( "mma_analyze_tagged_line: SUBMITTED: %ld...%ld", first_num, last_num );
    } else if( first_num > 0 && last_num >= first_num ) {
      presult->mma_first = first_num;
      presult->mma_last = last_num;
      trace( "mma_analyze_tagged_line: EVALUATED: %ld...%ld", first_num, last_num );
      type = MMA_TAG_EVALUATED;
    } else if( first_num == 0 ) {
      presult->vim_first = presult->vim_last = 0;
      type = MMA_TAG_VIRGIN;
      trace( "mma_analyze_tagged_line: VIRGIN" );
    } else {
      return MMA_TAG_MALFORMED;
    }
  }

  if( *lp == '/' ) {
    presult->options = lp;
    if( ! ( lp = (char_u *)strchr( (char *)lp, ']' ) ) )
      return MMA_TAG_MALFORMED;
  } else {
    presult->options = (char_u *)"";
  }

  if( *lp++ != ']' )
    return MMA_TAG_MALFORMED;

  trace( "  found `]'" );

  if( ! *lp )
    presult->payload = lp;
  else if( vim_iswhite( *lp++ ) )
    presult->payload = lp;
  else
    return MMA_TAG_MALFORMED;

  trace( "mma_analyze_tagged_line: returning %d", type );

  return type;
}


  /*
   * get_payload:
   *  Returns the pointer to the payload of a tagged line.
   *  This is a fast short cut which does no careful syntax checks, so
   *  you should only call it for well-formed tagged lines, or else
   *  the result is undefined (and may be NULL).
   */
  static char_u *
get_payload( lp )
  char_u *lp;
{
  lp = (char_u *)strchr( (char *)lp, ']' );
  if( ! lp ) return NULL;
  if( ! * ++lp )
    return lp;
  else if( vim_iswhite( *lp++ ) )
    return lp;
  else
    return NULL;
}

/*
 *  check whether the given tagged line contains an input-tag:
 */
  static int
contains_a_input_tag( lp )
  char_u *lp;
{
  while( vim_iswhite( *lp ) )
    ++lp;
  return  *lp == 'i' && lp[1] == '[';
}

/*
 *  check whether the given tagged line contains an include-tag:
 */
  static int
contains_a_include_tag( lp )
  char_u *lp;
{
  while( vim_iswhite( *lp ) )
    ++lp;
  return ! strncmp( (char *)lp, "include[", 8 );
}


/*
 *
 *   functions to navigate through and modify lines in an input file:
 *
 */


/*
 *  This will return the last line number of a block of lines starting at lnum
 *  and indented more than the given argument.
 *  Any trailing option lines are skipped, too.
 *  If line lnum+1 already fails this criterion, or doesn't even exist, it just
 *  returns lnum, with no checks performed on this line.
 *  Comments inside the block are allowed and skipped.
 */
  static linenr_t
mma_skip_block( buf, lnum, indent )
  BUF *buf;
  linenr_t lnum;
  colnr_t indent;
{
  char_u *lp;

  /*  skip lines of the block  */
  for(;;) {
    lp = child_get_line( buf, ++lnum );
    if( ! lp )
      break;
    if( get_indent_str( lp ) <= indent )
      if( ! mma_is_comment_line( lp ) )
        break;
  }
  /*  skip trailing option lines, if any  */
  for(;;) {
    if( ! lp )
      break;
    if( ! mma_is_option_line( lp ) )
      break;
    do {
      lp = child_get_line( buf, ++lnum );
    } while( mma_is_comment_line( lp ) );
  }
  return lnum - 1;
}


/*
 *  Replace the tag in the specified line in the buffer by a new one.
 *  Different lengths are taken care of and indentation is preserved.
 *  The new tag, however, must not contain any indentation, and it must
 *  contain a full tag, up to but excluding the whitespace which separates
 *  the tag from its payload.
 */
  static int
mma_replace_tag( buf, lnum, newtag )
  BUF *buf;
  linenr_t lnum;
  char_u * newtag;
{
  char_u *oldline;
  char_u *newline;
  char_u *payload;
  colnr_t taglen;
  colnr_t indent;
  colnr_t payloadlen;
  char_u *pc;
  int rv;

  oldline = child_get_line( buf, lnum );
  if( ! oldline ) {
    EMSG( "mma_replace_tag: line does not exist" );
    return FAIL;
  }
  if( ! ( payload = get_payload( oldline ) ) ) {
    EMSG( "mma_replace_tag: not a tagged line" );
    return FAIL;
  }
  taglen = strlen( (char *)newtag );
  indent = get_indent_str( oldline );
  payloadlen = strlen( (char *)payload );
  if( ! ( newline = (char_u *)malloc( indent + taglen + 1 + payloadlen + 1 ) ) ) {
    EMSG( "mma_replace_tag: malloc() failed" );
    return FAIL;
  }
  for( pc = newline + indent - 1; pc >= newline; --pc )
    *pc = ' ';
  strcpy( (char *)newline + indent, (char *)newtag );
  newline[ indent + taglen ] = ' ';
  strcpy( (char *)newline + indent + taglen + 1, (char *)payload );
  rv = child_set_line( buf, lnum, newline );
  free( newline );
  return rv;
}


  static long int
mma_get_new_negative_tag( buf )
  BUF *buf;
{
  return buf->b_new_negative_tag_number--;
}


/*
 *  Choose a new negative tag number for the given input line which
 *  is unique within the specified buffer.
 *  The tagged line is replaced in the buffer and the new number
 *  is returned.
 */
  static long int
make_input_tag_unique( buf, lnum )
  BUF *buf;
  linenr_t lnum;
{
  long int newnum;
  char_u newtag[MAX_TAG_LEN];

  newnum = mma_get_new_negative_tag( buf );
  snprintf( newtag, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN ) "i[%ld]", newnum );
  mma_replace_tag( buf, lnum, newtag );
  return newnum;
}


/*
 *  This will delete the block headed by the tagged line lnum,
 *  including the tagged line itself.
 *  It will not touch trailing option lines, if any, but will delete any
 *  comment lines which are part of the block.
 *  It is not an error if the specified line does not exist.
 */
  int
mma_delete_block( buf, lnum )
  BUF *buf;
  linenr_t lnum;
{
  char_u *lp;
  colnr_t indent;

  trace( "mma_delete_block: called with lnum = %ld", (long int)lnum );
  if( ! ( lp = child_get_line( buf, lnum ) ) )
    return OK;
  indent = get_indent_str( lp );
  for(;;) {
    if( ! ( lp = child_get_line( buf, lnum + 1 ) ) )
      break;
    if( get_indent_str( lp ) > indent || mma_is_comment_line( lp ) ) {
      trace( "mma_delete_block: deleting: %d: >>>%s<<<", lnum, lp );
      if( ! child_delete_line( buf, lnum + 1 ) )
        return FAIL;
    } else {
      break;
    }
  }
  return child_delete_line( buf, lnum );
}


/*
 *  This should be called for a tagged input or include line and will destroy all
 *  corresponding output in the same cell.
 */
  int
mma_delete_corresponding_output( buf, lnum )
  BUF *buf;
  linenr_t lnum;
{
  char_u *lp;
  struct mma_tline_s result;

  trace( "mma_delete_corresponding_output: called with lnum = %ld", (long int)lnum );

  if( ! ( lp = child_get_line( buf, lnum ) ) ) {
    EMSG( "mma_delete_corresponding_output: invalid line number" );
    return FAIL;
  }
  if( mma_analyze_tagged_line( lp, & result ) == MMA_TAG_MALFORMED ) {
    EMSG( "mma_delete_corresponding_output: not a well-formed tagged line" );
    return FAIL;
  }
  if( ! contains_a_input_tag( lp ) && ! contains_a_include_tag( lp ) ) {
    EMSG( "mma_delete_corresponding_output: line contains neither input nor include tag" );
    return FAIL;
  }

  lnum = mma_skip_block( buf, lnum, get_indent_str(lp) );

  ++lnum;
  do {
    if( ! ( lp = child_get_line( buf, lnum ) ) )
      return TRUE;
    if( mma_is_separator_line( lp ) )
      return TRUE;
    if( mma_analyze_tagged_line( lp, & result ) != MMA_TAG_MALFORMED ) {
      if( contains_a_input_tag( lp ) || contains_a_include_tag( lp ) )
        return TRUE;
      else
        mma_delete_block( buf, lnum );
    } else {
      ++lnum;
    }
  } while( TRUE );
}


  static linenr_t
mma_find_end_of_output( buf, lnum )
  BUF *buf;
  linenr_t lnum;
{
  char_u *lp;
  struct mma_tline_s result;
  linenr_t last_match;

  trace( "mma_find_end_of_output: called with lnum = %ld", (long int)lnum );
  if( ! ( lp = child_get_line( buf, lnum ) ) ) {
    EMSG( "mma_delete_corresponding_output: invalid line number" );
    return FAIL;
  }
  if( mma_analyze_tagged_line( lp, & result ) == MMA_TAG_MALFORMED ) {
    EMSG( "mma_delete_corresponding_output: not a well-formed tagged line" );
    return FAIL;
  }
  if( ! contains_a_input_tag( lp ) && ! contains_a_include_tag( lp ) ) {
    EMSG( "mma_delete_corresponding_output: line contains neither input nor include tag" );
    return FAIL;
  }

  last_match = lnum = mma_skip_block( buf, lnum, get_indent_str( lp ) );

  while( TRUE ) {
    if( ! ( lp = child_get_line( buf, ++lnum ) ) )
      return last_match;
    if( mma_is_separator_line( lp ) )
      return last_match;
    if( mma_analyze_tagged_line( lp, & result ) != MMA_TAG_MALFORMED ) {
      if( contains_a_input_tag( lp ) || contains_a_include_tag( lp ) )
        return last_match;
      else
        last_match = lnum = mma_skip_block( buf, lnum, get_indent_str( lp ) );
    }
  }

  return last_match;
}



/*
 *
 *   functions sending input to Mathematica
 *
 */


/*
 *  Write a line synchronously to stdin of a child process.
 *  Proper tags will be prepended and the line will be terminated by
 *  a newline. the length of the tag is specified separately, so the
 *  tag need not be NUL-terminated and can be part of some other string.
 *  lp should not contain a tag but must point to the payload.
 */
  static int
write_input_line_to_mma( pc, lp, tag, taglen )
  CHILD *pc;
  char_u *lp;
  char_u *tag;
  colnr_t taglen;
{
  colnr_t llen;

  trace( "write_input_line_to_mma: tag: >>>%s<<< payload: >>>%s<<<", tag, lp );
  if( pc->c_in < 0 ) {
    EMSG( "write_input_line_to_mma: stdin not available (disconnected after error?)" );
    return FAIL;
  }
  if( ! ( llen = strlen( (char *)lp ) ) )
    return OK;
  if( write( pc->c_in, tag, taglen ) < taglen ) {
    EMSG( "write_input_line_to_mma: write() #1 failed, disconnecting stdin" );
    close( pc->c_in );
    pc->c_in = -1;
    return FAIL;
  }
  /*  protocol says: the tag ends with a ']' and must be followed by a ' ':  */
  if( write( pc->c_in, " ", 1 ) < 1 ) {
    EMSG( "write_input_line_to_mma: write() #2 failed, disconnecting stdin" );
    close( pc->c_in );
    pc->c_in = -1;
    return FAIL;
  }
  if( write( pc->c_in, lp, llen ) < llen ) {
    EMSG( "write_input_line_to_mma: write() #3 failed, disconnecting stdin" );
    close( pc->c_in );
    pc->c_in = -1;
    return FAIL;
  }
  if( write( pc->c_in, "\n", 1 ) < 1 ) {
    EMSG( "write_input_line_to_mma: write() #4 failed, disconnecting stdin" );
    close( pc->c_in );
    pc->c_in = -1;
    return FAIL;
  }
  return OK;
}

/*
 * write_option_line_to_mma:
 *  Option lines are transferred literally, after discarding any leading whitespace.
 *  The first character should then be a dash.
 */
  static int
write_option_line_to_mma( pc, lp )
  CHILD *pc;
  char_u *lp;
{
  size_t len;
  trace( "write_option_line_to_mma: >>>%s<<<", lp );
  if( pc->c_in < 0 ) {
    EMSG( "write_input_line_to_mma: stdin not available (disconnected after error?)" );
    return FAIL;
  }
  while( vim_iswhite( *lp ) )
    ++lp;
  len = strlen( (char *)lp );
  if( write( pc->c_in, lp, len ) != len ) {
    EMSG( "write_option_line_to_mma: write() #1 failed, disconnecting stdin" );
    close( pc->c_in );
    pc->c_in = -1;
    return FAIL;
  }
  if( write( pc->c_in, "\n", 1 ) < 1 ) {
    EMSG( "write_option_line_to_mma: write() #2 failed, disconnecting stdin" );
    close( pc->c_in );
    pc->c_in = -1;
    return FAIL;
  }
  return OK;
}

/*
 *  Submit a block of input lines to a child's stdin.
 *  The line number passed should be the block leader.
 *  Actually, this function will:
 *    - submit the input block itself, prepending the input tag
 *      to all lines;
 *    - submit all option lines following the input block;
 *    - submit the meta-line "SUBMIT\n", which will cause the input
 *      block to be submitted to Mma for evaluation.
 */
  static int
write_group_to_mma( buf, lnum, pc )
  BUF *buf;
  linenr_t lnum;
  CHILD *pc;
{
  char_u *lp, *ptag, *payload;
  char_u tag[ MAX_TAG_LEN ];
  struct mma_tline_s result;
  colnr_t taglen;
  colnr_t indent;

  trace( "write_group_to_mma: called with lnum = %d", lnum );

  if( ! ( lp = child_get_line( buf, lnum ) ) ) {
    EMSG( "write_group_to_mma: invalid line number" );
    return FAIL;
  }
  if( mma_analyze_tagged_line( lp, & result ) == MMA_TAG_MALFORMED ) {
    EMSG( "write_group_to_mma: not a well-formed tagged line" );
    return FAIL;
  }
  indent = get_indent_str( lp );
  for( ptag = lp; vim_iswhite(*ptag); ++ptag )
    ;
  taglen = ( (char_u *)strchr( (char *)ptag, ']' ) - ptag ) + 1;
  if( taglen >= MAX_TAG_LEN ) {
    EMSG( "tag too long" );
    return FAIL;
  }
  /*  we have to copy the tag as ptag may be invalidaded by
   *  the next call of child_get_line:
   */
  memcpy( tag, ptag, taglen );
  tag[ taglen ] = NUL;

  /*
   *  the block leader may already contain input, but if not, we
   *  send a space so that Mma will see at least _some_ input:
   */
  if( ! write_input_line_to_mma(
    pc, *result.payload ? result.payload : (char_u *)" ", tag, taglen ) ) {
    EMSG( "write_group_to_mma: error while writing input to child" );
    return FAIL;
  }

  do {
    if( ! ( lp = child_get_line( buf, ++lnum ) ) )
      break;
    if( get_indent_str( lp ) <= indent ) {
      if( mma_is_comment_line( lp ) )
        continue;
      else
        break;
    }
    for( payload = lp; vim_iswhite( *payload ); ++payload )
      ;
    if( ! write_input_line_to_mma( pc, payload, tag, taglen ) )
      return FAIL;
  } while( TRUE );

  while( lp && mma_is_option_line( lp ) ) {
    if( ! write_option_line_to_mma( pc, lp ) )
      return FAIL;
    do {
      lp = child_get_line( buf, ++lnum );
    } while( mma_is_comment_line( lp ) );
  }

  if( write( pc->c_in, "SUBMIT\n", 7 ) < 7 ) {
    EMSG( "write_group_to_mma: write() error when writing to child process" );
    return FAIL;
  }
  return OK;
}


/*
 *  mma_reset:
 *   To be called when attaching a fresh mma kernel to the buffer:
 *   it will reset all tag numbers to 0.
 */
  int
mma_reset( buf )
  BUF *buf;
{
  linenr_t lnum;
  char_u *lp;
  char_u newtag[MAX_TAG_LEN];
  char_u *s1, *s2;
  colnr_t indent;
  struct mma_tline_s result;

  buf->b_new_negative_tag_number = -1;

  for( lnum = 1; lnum <= buf->b_ml.ml_line_count; ++lnum ) {
    if( ! ( lp = child_get_line( buf, lnum ) ) ) {
      EMSG( "mma_reset: child_get_line() failed" );
      return FAIL;
    }

    indent = get_indent_str( lp );
    while( vim_iswhite( *lp ) )
      ++lp;

    if( mma_analyze_tagged_line( lp, & result ) != MMA_TAG_MALFORMED ) {

      for( s1 = newtag, s2 = lp; s1 < newtag + ( MAX_TAG_LEN - 4 ) ; ++s1, ++s2 )
        if( isalnum( *s2 ) || *s2 == '_' )
          *s1 = *s2;
        else
          break;
      strcpy( (char *)s1, "[0]" );
      if( ! mma_replace_tag( buf, lnum, newtag ) ) {
        EMSG( "mma_reset: replace_tag() failed" );
        return FAIL;
      }
      lnum = mma_skip_block( buf, lnum, indent );
    }
  }
  return OK;
}


/*
 *
 *  functions evaluating appropriate parts of a file
 *
 */


/*
 *  mma_eval_file:
 *    Read a file and submit all input blocks to Mma for evaluation.
 *    Recursive include statements are allowed.
 *    The file will not be modified, but all resulting output will be
 *    directed into the original buffer (this can of course be suppressed 
 *    by a suitable option).
 *    Unique tag numbers are requested from the specified buffer, and the 
 *    range of numbers used up is returned.
 */

/*
 *   We first define a macro to read a new line of input into a
 *   dynamically growing buffer, unless there is still input left
 *   in the buffer:
 */

#define MMA_FGETS \
  if( bp == lp )   /*  ie buffer is empty  */  { \
    eol = FALSE;   \
    while( ! eol ) {   \
      if( bp - lp >= bufsize - 2 ) {   \
        if( ! ( newp = (char_u *)realloc( lp, bufsize *= 2 ) ) ) {   \
          EMSG( "MMA_FGETS: realloc() failed" );   \
          free( lp );   \
          fclose( stream );   \
          return FAIL;   \
        }   \
        bp += ( newp - lp );   \
      }   \
      switch( c = fgetc( stream ) ) {   \
        case EOF:   \
          eol = eof = TRUE;   \
          break;   \
        case '\n':   \
          eol = TRUE;   \
          break;   \
        default:   \
          *bp++ = c;   \
          break;   \
      }   \
    }   \
    *bp = NUL;   \
  }

struct mma_linebuf_s {
  char_u *line;
  struct mma_linebuf_s *next, *prev;
};

  static int
mma_eval_file( buf, path, pfirst, plast, argopts )
  BUF *buf;           /*  used only to retrieve unique tag numbers  */
  char_u *path;       /*  path of file to be evaluated  */
  long int *pfirst;   /*  first tag number used (0 iff none)  */
  long int *plast;    /*  last tag number used (0 iff none)  */
  struct mma_linebuf_s *argopts;
{
  CHILD *pc;
  FILE *stream;
  char_u tag[ MAX_TAG_LEN ];
  colnr_t taglen;
  char_u *lp, *bp, *newp;
  size_t bufsize = 1000;
  int c;
  int eof, eol;
  colnr_t indent;
  long int new_first, new_last;
  struct mma_tline_s result;
  int rv;
  char_u *recursive_path;
  struct mma_linebuf_s *opts_root, *locopts_last, *popt, *popt_tmp;

  stream = NULL;
  lp = NULL;
  recursive_path = NULL;
  *pfirst = *plast = 0;
  opts_root = argopts;
  locopts_last = NULL;
  rv = FAIL;

  trace( "mma_eval_file: >>>%s<<<", path );
  if( ! opts_root ) trace( "  no options" );
  for( popt = opts_root; popt; popt = popt->next )
    trace( "  option line: >>>%s<<<", popt->line );

  if( ! ( pc = find_child_by_name( buf, "mma" ) ) ) {
    EMSG( "mma_eval_file: no mathematica child process found" );
    goto bailout;
  }
  if( pc->c_in < 0 ) {
    EMSG( "mma_eval_file: stdin of mma kernel not available (closed after error?)" );
    goto bailout;
  }

  if( ! ( stream = fopen( path, "r" ) ) ) {
    EMSG( "mma_eval_file: fopen() failed" );
    goto bailout;
  }
  if( ! ( lp = (char_u *)malloc( bufsize ) ) ) {
    EMSG( "mma_eval_file: malloc() failed" );
    fclose( stream );
    goto bailout;
  }

  eof = FALSE;
  bp = lp;
  while( ! eof || bp > lp ) {

    MMA_FGETS
    trace( "mma_eval_file: got line: >>>%s<<<", lp );
    if( mma_analyze_tagged_line( lp, & result ) != MMA_TAG_MALFORMED ) {

      if( contains_a_input_tag( lp ) ) {

        indent = get_indent_str( lp );
        *plast = mma_get_new_negative_tag( buf );
        if( ! *pfirst ) *pfirst = *plast;
        snprintf( tag, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN ) "i[%ld]", *plast );
        taglen = strlen( tag );
        if( ! write_input_line_to_mma(
          pc, *result.payload ? result.payload : (char_u *)" ", tag, taglen
        ) )
          goto bailout;
        bp = lp;

        while( ! eof ) {
          MMA_FGETS
          if( get_indent_str( lp ) > indent ) {
            for( bp = lp; vim_iswhite( *bp ); ++bp )
              ;
            if( ! write_input_line_to_mma( pc, bp, tag, taglen ) )
              goto bailout;
            bp = lp;
          } else {
            if( mma_is_comment_line( lp ) )
              bp = lp;
            else
              break;
          }
        }
        while( mma_is_option_line( lp ) ) {
          if( ! write_option_line_to_mma( pc, lp ) )
            goto bailout;
          do {
            bp = lp;
            MMA_FGETS
          } while( mma_is_comment_line( lp ) );
        }
        for( popt = argopts; popt; popt = popt->next )
          if( ! write_option_line_to_mma( pc, popt->line ) )
            goto bailout;

        if( write( pc->c_in, "SUBMIT\n", 7 ) < 7 ) {
          EMSG( "mma_eval_file: write() error when writing to child process" );
          goto bailout;
        }

        /*  here, we may leave with bp > lp, so we parse this line in the next  */
        /*  iteration of this function's main loop!                             */

      } else if( contains_a_include_tag( lp) ) {

        if( ! ( recursive_path = (char_u *)strdup( get_payload( lp ) ) ) ) {
          EMSG( "mma_eval_file: malloc() failed for file name" );
          goto bailout;
        }
        bp = lp;
        while( TRUE ) {
          while( TRUE ) {
            MMA_FGETS;
            if( mma_is_comment_line( lp ) )
              bp = lp;
            else
              break;
          }
          if( mma_is_option_line( lp ) ) {
            if( ! ( popt = (struct mma_linebuf_s *)malloc( sizeof( struct mma_linebuf_s ) ) ) ) {
              EMSG( "mma_eval_file: malloc() failed" );
              goto bailout;
            }
            if( ! ( popt->line = (char_u *)strdup( lp ) ) ) {
              EMSG( "mma_eval_file: malloc() failed" );
              free( popt );
              goto bailout;
            }

            if( opts_root == argopts ) opts_root = popt;
            popt->next = argopts;
            if( argopts ) argopts->prev = popt;
            if( locopts_last )
              popt->prev = locopts_last, locopts_last->next = popt;
            else
              popt->prev = NULL;
            locopts_last = popt;
            bp = lp;
          } else {
            break;
            /*  here, we may leave with bp > lp, so we parse this line in the next  */
            /*  iteration of this function's main loop!                             */
          }
        }

        trace( "mma_eval_file: will call myself with path >>>%s<<<", path );
        if( ! mma_eval_file( buf, recursive_path, &new_first, &new_last, opts_root ) ) {
          EMSG( "mma_eval_file: recursive call to mma_eval_file() failed" );
          goto bailout;
        }
        if( ! *pfirst ) *pfirst = new_first;
        if( new_last ) *plast = new_last;

        free( recursive_path ), recursive_path = NULL;
        popt = opts_root;
        while( popt != argopts ) {
          free( popt->line );
          popt_tmp = popt->next;
          free( popt );
          popt = popt_tmp;
        }
        if( argopts ) argopts->prev = NULL;
        opts_root = argopts;
        locopts_last = NULL;

      } else {   /*  tagged, but neither input nor include  */
        /*  nop, just skip this line  */
        bp = lp;
      }

    } else {   /*  no tag at all  */
      /*  nop, just skip this line  */
      bp = lp;
    }
  }

  rv = OK;

  bailout:

  /*  having destructors would make life so much simpler...  */
  if( lp ) free( lp );
  free( recursive_path ), recursive_path = NULL;
  if( stream ) fclose( stream );
  popt = opts_root;
  while( popt != argopts ) {
    free( popt->line );
    popt_tmp = popt->next;
    free( popt );
    popt = popt_tmp;
  }
  if( argopts ) argopts->prev = NULL;

  trace( 
    "mma_eval_file: returning rv = %d, first = %ld, last = %ld"
  , rv, *pfirst, *plast
  );

  return rv;
}



/*
 *  A compare function to be used in the qsort() library function:
 */
  static int
mma_compare_linenr_t( pi1, pi2 )
  void *pi1;
  void *pi2;
{
  if( *(linenr_t *)pi1 < *(linenr_t *)pi2 )
    return -1;
  else if( *(linenr_t *)pi1 > *(linenr_t *)pi2 )
    return 1;
  else
    return 0;
}

/*
 *  evaluate a list of input blocks:
 *    pi should point to an array of line numbers of input block leaders,
 *    terminated by a negative entry. the array need not be sorted, and
 *    lines will be evaluated in the given order. however, array entries
 *    should better be unique.
 *
 *    We first submit all input to Mma in the given order; we then sort the
 *    array and delete any existing output in order of ascending line numbers.
 */
  static int
mma_eval_list( buf, pi )
  BUF *buf;
  linenr_t *pi;
{
  CHILD *pc;
  int j;
  int rv;
  char_u *lp;
  char_u *path;
  linenr_t lnum;
  long int first_tagnum, last_tagnum;
  char_u tag[ MAX_TAG_LEN ];
  struct mma_linebuf_s *opts_root, *opts_last, *popt, *popt_tmp;
  struct mma_tline_s result;

  if( ! *pi ) return OK;

  rv = OK;
  if( ! ( pc = find_child_by_name( buf, "mma" ) ) ) {
    EMSG( "mma_eval_list: no mathematica child process found" );
    return FAIL;
  }
  if( pc->c_in < 0 ) {
    EMSG( "mma_eval_list: mma kernel's stdin not available (disconnected after error?)" );
    return FAIL;
  }

  for( j = 0; pi[j] > 0; ++j ) {

    if( ! ( lp = child_get_line( buf, pi[j] ) ) ) {
      EMSG( "mma_eval_list: child_get_line() failed" );
      return FAIL;
    }
    if( mma_analyze_tagged_line( lp, & result ) == MMA_TAG_MALFORMED ) {
      EMSG( "mma_eval_list: parse error: tagged line not well-formed" );
      return FAIL;
    }

    if( contains_a_input_tag( lp ) ) {

      if( ! make_input_tag_unique( buf, pi[j] ) ) {
        EMSG( "mma_eval_list: make_input_tag_unique() failed" );
        rv = FAIL;
      } else if( ! write_group_to_mma( buf, pi[j], pc ) ) {
        EMSG( "mma_eval_list: write_group_to_mma() failed" );
        rv = FAIL;
      }

    } else if( contains_a_include_tag( lp ) ) {

      if( ! ( path = (char_u *)strdup( (char *)(result.payload) ) ) ) {
        EMSG( "mma_eval_list: strdup() failed for include file name" );
        rv = FAIL;
        continue;
      }
      opts_root = opts_last = NULL;
      lnum = pi[j];
      while( TRUE ) {
        do {
          lp = child_get_line( buf, ++lnum );
        } while( mma_is_comment_line( lp ) );
        if( mma_is_option_line( lp ) ) {

          if( ! ( popt = (struct mma_linebuf_s *)malloc( sizeof( struct mma_linebuf_s ) ) ) ) {
            EMSG( "mma_eval_list: malloc() failed" );
            rv = FAIL;
            break;
          }
          if( ! ( popt->line = (char_u *)strdup( (char *)lp ) ) ) {
            EMSG( "mma_eval_list: malloc() failed" );
            free( popt );
            rv = FAIL;
            break;
          }
          popt->next = NULL;
          if( ! opts_root ) 
            opts_root = popt, popt->prev = NULL;
          else
            opts_last->next = popt, popt->prev = opts_last;
          opts_last = popt;

        } else {
          break;
        }
      }

      rv = mma_eval_file( buf, path, &first_tagnum, &last_tagnum, opts_root );
      free( path );
      popt = opts_root;
      while( popt ) {
        free( popt->line );
        popt_tmp = popt->next;
        free( popt );
        popt = popt_tmp;
      }

      snprintf( 
        tag, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN )
        first_tagnum == last_tagnum ? "include[%ld]" : "include[%ld...%ld]"
      , first_tagnum, last_tagnum 
      );
      mma_replace_tag( buf, pi[j], tag );

    } else {
      EMSG( "mma_eval_list: neither input nor include line" );
      return FAIL;
    }

  }

  /*  before we can delete the existing output, we sort the line numbers
   *  and then delete in descending order, so none of the line numbers will
   *  change before we get to it:
   */

  qsort( pi, j, sizeof( linenr_t ), &mma_compare_linenr_t );

  for( --j ; j >= 0; --j ) {
    if( ! mma_delete_corresponding_output( buf, pi[j] ) ) {
      EMSG( "mma_eval_list: invalid line number" );
      rv = FAIL;
    }
  }

  return rv;
}


  int
mma_eval_range( buf, from, to )
  BUF *buf;
  linenr_t from;
  linenr_t to;
{
  linenr_t *pi;
  int pisize;
  linenr_t *newpi;

  linenr_t lnum;
  char_u *lp;
  int rv;
  int index;
  struct mma_tline_s result;

  pisize = 64;
  if( ! ( pi = (linenr_t *)malloc( pisize * sizeof( linenr_t ) ) ) ) {
    EMSG( "mma_eval_range: malloc() failed" );
    return FAIL;
  }

  rv = OK;
  index = 0;
  for( lnum = from; lnum <= to; ++lnum ) {
    if( ! ( lp = child_get_line( buf, lnum ) ) ) {
      EMSG( "mma_eval_range: invalid line number" );
      rv = FAIL;
      continue;
    }
    if( mma_analyze_tagged_line( lp, & result ) == MMA_TAG_MALFORMED )
      continue;
    if( contains_a_input_tag( lp ) || contains_a_include_tag( lp ) ) {
      if( index >= pisize - 2 ) {    /*  keep one entry on reserve  */
        if( ! ( newpi = (linenr_t *)realloc( pi, (pisize *= 2) * sizeof(linenr_t) ) ) ) {
          EMSG( "mma_eval_range: malloc() failed" );
          free( pi );
          return FAIL;
        }
        pi = newpi;
      }
      pi[ index++ ] = lnum;

      /*  there may be more lines _looking_ like input lines in the block,
       *  so we have to skip them:
       */
      lnum = mma_skip_block( buf, lnum, get_indent_str(lp) );
    }
  }
  pi[ index ] = 0;

  if( ! mma_eval_list( buf, pi ) )
    rv = FAIL;

  if( pi ) free( pi );
  return rv;
}


/*
 *  some special commands to interrupt ongoing computations:
 *
 *    This is currently just an ``emergency exit'', it won't reset the
 *    negative tag numbers.
 *    These functions have f_unction counterparts in eval.c.
 *
 */

  int
mma_abort( buf )
  BUF *buf;
{
  CHILD *pc;

  if( ! ( pc = find_child_by_name( buf, "mma" ) ) ) {
    EMSG( "mma_abort: no mathematica child process found" );
    return FAIL;
  }
  if( write( pc->c_in, "ABORT\n", 6 ) < 6 ) {
    EMSG( "mma_abort: write() error when writing to child process" );
    return FAIL;
  }
  return OK;
}


  int
mma_interrupt( buf )
  BUF *buf;
{
  CHILD *pc;

  if( ! ( pc = find_child_by_name( buf, "mma" ) ) ) {
    EMSG( "mma_interrupt: no mathematica child process found" );
    return FAIL;
  }
  if( write( pc->c_in, "INTERRUPT\n", 10 ) < 10 ) {
    EMSG( "mma_interrupt: write() error when writing to child process" );
    return FAIL;
  }
  return OK;
}


  int
mma_purge_input( buf )
  BUF *buf;
{
  CHILD *pc;

  if( ! ( pc = find_child_by_name( buf, "mma" ) ) ) {
    EMSG( "mma_purge_input: no mathematica child process found" );
    return FAIL;
  }
  if( write( pc->c_in, "PURGE INPUT\n", 12 ) < 12 ) {
    EMSG( "mma_purge_input: write() error when writing to child process" );
    return FAIL;
  }
  return OK;
}

  int
mma_purge_output( buf )
  BUF *buf;
{
  CHILD *pc;

  if( ! ( pc = find_child_by_name( buf, "mma" ) ) ) {
    EMSG( "mma_purge_output: no mathematica child process found" );
    return FAIL;
  }
  if( write( pc->c_in, "PURGE OUTPUT\n", 13 ) < 13 ) {
    EMSG( "mma_purge_output: write() error when writing to child process" );
    return FAIL;
  }
  return OK;
}



/*
 *
 *  Below this comment, we have functions dealing with the "reverse" task:
 *  reading and handling input from mma:
 *
 */


/*
 * mma_handle_meta:
 *  Handles special messages from ml2vim which don't contain actual output
 *  from mathematica.
 *
 */

  int
mma_handle_meta( pc, vim_tag_num, mma_tag_num, payload )
  CHILD * pc;
  long int vim_tag_num;
  long unsigned mma_tag_num;
  char_u *payload;
{
  BUF *buf;
  char_u *lp, *s;
  struct mma_tline_s result;
  char_u tag[ MAX_TAG_LEN ];
  linenr_t lnum;
  int type;

  buf = pc->c_buf;

  trace( "mma_handle_meta: %ld, %ld, >>>%s<<<", vim_tag_num, mma_tag_num, payload );

  if( ! strcmp( payload, "EOC" ) ) {

    /*  EOC: this is send whenever the Mma kernel has produced a new    */
    /*  input prompt, to indicate that evaluation of one cell has been  */
    /*  completed.                                                      */

    for( lnum = 1; lnum <= buf->b_ml.ml_line_count; ++lnum ) {
      if( ! ( lp = child_get_line( buf, lnum ) ) )
        return TRUE;
      while( vim_iswhite( *lp ) )
        ++lp;
      switch( ( type = mma_analyze_tagged_line( lp, & result ) ) ) {
        case MMA_TAG_MALFORMED:
        case MMA_TAG_EVALUATED:
        case MMA_TAG_VIRGIN:
          break;
        case MMA_TAG_SUBMITTED:
        case MMA_TAG_PARTIAL:
          if( contains_a_input_tag( lp ) || contains_a_include_tag( lp ) ) {
            if( result.vim_first >= vim_tag_num && vim_tag_num >= result.vim_last ) {
              /*  bingo!  */
              if( vim_tag_num == result.vim_last ) {
                /*  this seems to be fully evaluated, but we perform some sanity checks:  */
                if( type == MMA_TAG_PARTIAL ) {
                  if( mma_tag_num > result.mma_last + 1 || mma_tag_num < result.mma_last ) {
                    trace( 
                      "mma_handle_meta: EOC message out of sync:  %ld/%ld %ld/%ld got: %ld/%ld"
                    , result.vim_first, result.vim_last
                    , result.mma_first, result.mma_last
                    , vim_tag_num, mma_tag_num
                    );
                    return FAIL;
                  }
                }
                for( s = tag; *lp != '[' && s < tag + MAX_TAG_LEN; *s++ = *lp++ )
                  ;
                snprintf( 
                  s, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN - ( s - tag ) )
                  mma_tag_num > result.mma_first ?  "[%ld...%ld]" : "[%ld]"
                , result.mma_first, mma_tag_num
                );
                trace( "mma_handle_meta: replacing tag in line %d: >>>%s<<<", lnum, lp );
                trace( "  new tag: >>>%s<<<", tag );
                return mma_replace_tag( buf, lnum, tag );
              } else {
                /*  this is only partially evaluated, but may still need an update:  */
                if( type == MMA_TAG_SUBMITTED ||  mma_tag_num - result.mma_last == 1 ) {
                  for( s = tag; *lp != '[' && s < tag + MAX_TAG_LEN; *s++ = *lp++ )
                    ;
                  snprintf(
                    s, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN - ( s - tag ) )
                    "[%ld...%ld,%ld...%ld]"
                  , result.vim_first, result.vim_last
                  , type == MMA_TAG_SUBMITTED ? mma_tag_num : result.mma_first, mma_tag_num
                  );
                  trace( "mma_handle_meta: replacing tag in line %d: >>>%s<<<", lnum, lp );
                  trace( "  new tag: >>>%s<<<", tag );
                  return mma_replace_tag( buf, lnum, tag );
                } else if( mma_tag_num == result.mma_last ) {
                  return OK;   /*  ok: already up-to-date  */
                } else {
                  trace( 
                    "mma_handle_meta: EOC message out of sync:  %ld/%ld %ld/%ld got: %ld/%ld"
                  , result.vim_first, result.vim_last
                  , result.mma_first, result.mma_last
                  , vim_tag_num, mma_tag_num
                  );
                  return FAIL;
                }
              }
            }
          }
          break;
        default:
          return FAIL;
      }
    }

  } else if( ! strcmp( payload, "REFRESH" ) ) {

    trace( "mma_handle_meta: got REFRESH" );

    /*  REFRESH: demands a refresh of the display. How often this is  */
    /*  done depends on the setting of the refresh option.            */

    c_refresh_now = c_need_refresh;

  } else {

    /*  default is currently: ignore silently, except when debugging  */

    trace( "***** mma_handle_meta: can't handle: >>>%s<<<", payload );

  }

  return TRUE;
}


/*
 * handler for asynchronous input from mathematica:
 *
 *  This function handles a single NUL terminated input line, which should
 *  start with a proper tag (see mma_analyze_tagged_line), but must not
 *  contain number ranges.
 *
 *  The following tag names have a defined meaning:
 *    i: input lines
 *    o: mathematica output
 *    m: mathematica error messages
 *    d: graphics output (they will most likely contain postscript)
 *    t: text (can be caused eg by Print[])
 *    M: meta line (see mma_handle_meta()).
 *
 *  The tag numbers in brackets are used to group lines correctly:
 *  the first number is either entered by the user (should be 0) or issued by 
 *  vim (these are negative).
 *  The second number is issued by mma and will be positive.
 *
 *  The input line may be messed up by this function, so the caller may
 *  have to copy it.
 */

  int
handle_async_mma( pc, line ) 
  CHILD *pc;
  char_u *line;
{
  int rv;

  /*  the following are set when parsing the input:  */
  long int vim_tag_num;      /* non-positve number entered by user or issued by vim */
  long unsigned mma_tag_num; /* positive number issued by the mma kernel */
  char_u *lbracket;
  char_u *rbracket;
  char_u *payload, *payload_end;
  size_t payloadlen;

  /*  the following are set when parsing lines of the buffer:  */
  BUF *buf = pc->c_buf;
  linenr_t lnum;
  int type;
  char_u *lp, *ptag;
  int indent;
  struct mma_tline_s result;

  /*  the following are used when modifying the buffer  */
  char_u tag[MAX_TAG_LEN];
  int do_piggyback_output;
  int line_is_handled;
  char_u *newline;
  linenr_t corresponding_in_line;
  linenr_t matching_orphan_line;
  int have_passed_corresponding_cell;

  trace( "handle_async_mma: got >>>%s<<<", line );

  rv = FAIL;
  newline = NULL;

  /*  first, do some sanity checks and extract the tag numbers:  */

  if( ! * line ) return OK;      /* empty line: do nothing */

  if( ! ( lbracket = (char_u *)strchr( (char *)line, '[' ) ) ) {
    EMSG( "handle_async_mma: missing `[' in input line" );
    return FAIL;
  }
  if( ! ( rbracket = (char_u *)strchr( (char *)lbracket, ']' ) ) ) {
    EMSG( "handle_async_mma: missing `]' in input line" );
    return FAIL;
  }

  if( rbracket[1] != ' ' ) {
    EMSG( "handle_async_mma: parse error in input line" );
    return FAIL;
  }

  *lbracket = NUL;     /* isolate the tag name */

  payload = rbracket + 2;
  payload_end = payload + ( payloadlen = strlen( (char *)payload ) );

  rv = sscanf( (char *)lbracket + 1, "%ld,%lu", &vim_tag_num, &mma_tag_num );
  if( rv < 2 ) {
    EMSG( "handle_async_mma: sscanf() failed for input line" );
    return FAIL;
  }

  if( line[0] == 'M' && ! line[1] ) {
    return mma_handle_meta( pc, vim_tag_num, mma_tag_num, payload );
  }

  /*
   *  tags may contain flags after the second number before the rbracket.
   *  currently, only one flag is evaluated, all others are silently
   *  ignored (note, however, that -0 is _defined_ as a NOP).
   *  This should at some time be done more systematically.
   */
  do_piggyback_output = ( strstr( (char *)lbracket + 1, "/p" ) != NULL );

  c_need_refresh = TRUE;

  /*   find the appropriate position to insert the line into the buffer:  */

  line_is_handled = have_passed_corresponding_cell = FALSE;
  corresponding_in_line = matching_orphan_line = 0;
  lnum = 0;
  while( 
       ++lnum <= buf->b_ml.ml_line_count 
    && ! line_is_handled 
    && ! have_passed_corresponding_cell
  ) {

    if( ! ( lp = child_get_line( buf, lnum ) ) ) {
      EMSG( "mma_handle_async: child_get_line() failed" );
      goto bailout;
    }
    for( ptag = lp; vim_iswhite( *ptag ); ++ptag )
      ;
    switch( type = mma_analyze_tagged_line( ptag, & result ) ) {

      case MMA_TAG_MALFORMED:

        /*  this is not a ordinary tagged line, but it might contain an orphan:  */
        if( ! strncmp( (char *)ptag, "orphan: ", 8 ) ) {
          ptag += 8;
          if( mma_analyze_tagged_line( ptag, & result ) == MMA_TAG_PARTIAL ) {
            if( ! strncmp( (char *)ptag, (char *)line, lbracket - line ) ) {
              if( result.mma_first == mma_tag_num && result.vim_first == vim_tag_num ) {
                trace( 
                  "handle_async_mma: found existing orphan block at line %ld"
                , (long int)lnum 
                );
                matching_orphan_line = lnum;
              }
            }
          }
        }
        break;

      case MMA_TAG_VIRGIN:
        break;

      case MMA_TAG_SUBMITTED:
      case MMA_TAG_PARTIAL:
        /*  we might need to update the tag, and maybe create a new output block here  */

        if( ! contains_a_input_tag( ptag ) && ! contains_a_include_tag( ptag ) )
          break;

        trace( "handle_mma_async: found in line, type = %d: >>>%s<<<", type, ptag );
        trace( 
          "  result.vim_first/last: %ld/%ld  result.mma_first/last: %ld/%ld"
        , result.vim_first, result.vim_last
        , result.mma_first, result.mma_last
        );
        trace( "  vim_tag_num: %ld  mma_tag_num: %ld", vim_tag_num, mma_tag_num );
        if( vim_tag_num < result.vim_last || vim_tag_num > result.vim_first ) {
          if( corresponding_in_line > 0 )
            /*  a suitable cell exists and we passed it, so we can break out:  */
            have_passed_corresponding_cell = TRUE;
          break;
        }
        corresponding_in_line = lnum;
        trace( "  vim's numbers match!" );

        if( type == MMA_TAG_SUBMITTED || mma_tag_num == result.mma_last + 1 ) {

          trace( "  mma's number matches!" );
          /*  we need to update the tag  */

          if( result.vim_first == result.vim_last )
            snprintf( 
              tag, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN ) "%s[%ld,%ld]"
            , contains_a_input_tag( ptag ) ? "i" : "include"
            , vim_tag_num, mma_tag_num
            );
          else
            snprintf( 
              tag, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN ) "%s[%ld...%ld,%ld...%ld]"
            , contains_a_input_tag( ptag ) ? "i" : "include"
            , result.vim_first, result.vim_last
            , type == MMA_TAG_SUBMITTED ? mma_tag_num : result.mma_first
            , mma_tag_num
            );
          if( ! mma_replace_tag( buf, lnum, tag ) ) {
            EMSG( "handle_async_mma: mma_replace_tag() failed" );
            goto bailout;
          }
          trace( "handle_async_mma: updated tag in %d: >>>%s<<<", lnum, lp );
          trace( "  new tag: >>>%s<<<", tag );
        }
        break;

      case MMA_TAG_EVALUATED:
        /*  this might be the start of an output block which we can append to:  */
        if( ! strncmp( ptag, line, lbracket - line ) ) {
          if( mma_tag_num == result.mma_first ) {   /* bingo! */
            trace( "handle_async_mma: found existing output block: lnum = %ld", (long int)lnum );
            indent = get_indent_str( lp );
            lnum = mma_skip_block( buf, lnum, indent );
            indent += 2;
            if( ! ( newline = (char_u *)malloc( indent + payloadlen + 2 ) ) ) {
              EMSG( "handle_async_mma: malloc() failed" );
              goto bailout;
            }
            ptag = newline;
            while( indent-- > 0 )
              *ptag++ = ' ';

            strcpy( ptag, payload );
            if( ! child_append_line( buf, lnum, newline, strlen( newline ) + 1 ) ) {
              EMSG( "mma_handle_async: child_append_line() failed" );
              goto bailout;
            }
            line_is_handled = TRUE;
          }
        }
        break;

      default:
        EMSG( "mma_handle_async: internal error" );
        goto bailout;

    }
    trace( "  after switch" );
  }
  trace( "  after loop" );

  if( ! line_is_handled && corresponding_in_line > 0 ) {
    /*  we can create a new output block  */

    trace( "handle_async_mma: creating output block" );

    if( ! ( lp = child_get_line( buf, corresponding_in_line ) ) ) {
      EMSG( "mma_handle_async: child_get_line() failed" );
      goto bailout;
    }
    indent = get_indent_str( lp );
    lnum = mma_find_end_of_output( buf, corresponding_in_line );

    if( ! ( newline = (char_u *)malloc( indent + MAX_TAG_LEN + 1 + payloadlen + 1 ) ) ) {
      EMSG( "handle_async_mma: malloc() failed" );
      goto bailout;
    }
    for( lp = newline; lp < newline + indent; ++lp )
      *lp = ' ';
    snprintf(
      newline + indent, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN + 1 ) "%s[%ld]"
    , line, mma_tag_num
    );
    if( do_piggyback_output )
      strcat( (char *)newline, (char *)payload - 1 );
    if( ! child_append_line( buf, lnum, newline, strlen( newline ) + 1 ) ) {
      EMSG( "mma_handle_async: child_append_line() failed" );
      goto bailout;
    }
    if( ! do_piggyback_output ) {
      indent += 2;
      for( lp = newline; lp < newline + indent; ++lp )
        *lp = ' ';
      strcpy( (char *)newline + indent, (char *)payload );
      if( ! child_append_line( buf, lnum + 1, newline, strlen( newline ) + 1 ) ) {
        EMSG( "mma_handle_async: child_append_line() failed" );
        goto bailout;
      }
    }
    line_is_handled = TRUE;
  }

  if( ! line_is_handled && matching_orphan_line > 0 ) {
    /*  we can append to an existing orphan block  */
    trace( "handle_async_mma: line will be appended to existing orphan block" );
    if( ! ( lp = child_get_line( buf, matching_orphan_line ) ) ) {
      EMSG( "mma_handle_async: child_get_line() failed" );
      goto bailout;
    }
    indent = get_indent_str( lp );
    lnum = mma_skip_block( buf, lnum, indent );
    indent += 2;
    if( ! ( newline = (char_u *)malloc( indent + payloadlen + 1 ) ) ) {
      EMSG( "handle_async_mma: malloc() failed" );
      goto bailout;
    }
    ptag = newline;
    while( indent-- > 0 )
      *ptag++ = ' ';
    strcpy( ptag, payload );
    if( ! child_append_line( buf, lnum, newline, strlen( newline ) + 1 ) ) {
      EMSG( "mma_handle_async: child_append_line() failed" );
      goto bailout;
    }
    line_is_handled = TRUE;
  }

  if( ! line_is_handled ) {
    /*  couldn't find a place to insert line, so we create an orphan block:  */
    trace( "handle_async_mma: creating orphan block" );

    if( ! ( newline = (char_u *)malloc( MAX_TAG_LEN + payloadlen + 2 ) ) ) {
      EMSG( "handle_async_mma: malloc() failed" );
      goto bailout;
    }

    snprintf(
      newline, SNPRINTF_EXTRA_ARG( MAX_TAG_LEN + 1 ) "orphan: %s[%ld,%ld]"
    , line, vim_tag_num, mma_tag_num
    );
    if( do_piggyback_output )
      strcat( (char *)newline, (char *)payload - 1 );
    if( ! child_append_line( buf, buf->b_ml.ml_line_count, newline, strlen( newline ) + 1 ) ) {
      EMSG( "mma_handle_async: child_append_line() failed" );
      goto bailout;
    }
    if( ! do_piggyback_output ) {
      strcpy( (char *)newline, "  " );
      strcpy( (char *)newline + 2, payload );
      if( ! child_append_line( buf, buf->b_ml.ml_line_count, newline, strlen( newline ) + 1 ) ) {
        EMSG( "mma_handle_async: child_append_line() failed" );
        goto bailout;
      }
    }

    line_is_handled = TRUE;
  }

  rv = TRUE;

  bailout:

  trace( "  before free" );
  if( newline ) free( newline );

  trace( "  before returning %d", rv );
  return rv;
}

#endif  /* defined( USE_MMA ) */

